local ReprintsDeskServiceInternal = {};
ReprintsDeskServiceInternal.IsProduction = true;
ReprintsDeskServiceInternal.UserAgent = nil;
ReprintsDeskServiceInternal.Username = nil;
ReprintsDeskServiceInternal.Password = nil;

ReprintsDeskService = ReprintsDeskServiceInternal;

luanet.load_assembly("System");
luanet.load_assembly("System.Xml");

local client = nil;

local log = types["log4net.LogManager"].GetLogger(rootLogger .. ".ReprintsDeskService");

local types = {};
types["System.Uri"] = luanet.import_type("System.Uri");
types["System.Text.Encoding"] = luanet.import_type("System.Text.Encoding");
types["System.Net.WebClient"] = luanet.import_type("System.Net.WebClient");
types["System.Xml.XmlDocument"] = luanet.import_type("System.Xml.XmlDocument");
types["System.Xml.XmlNamespaceManager"] = luanet.import_type("System.Xml.XmlNamespaceManager");
types["System.Activator"] = luanet.import_type("System.Activator");
types["System.Reflection.Assembly"] = luanet.import_type("System.Reflection.Assembly");
types["StreamReader"] = luanet.import_type("System.IO.StreamReader");

local webClientAssembly = types["System.Reflection.Assembly"].LoadFile(AddonInfo.Directory .. "\\ILLiad-WebClient-Override.dll");

types["ILLiad-WebClient-Override.ILLiadWebClientOverride"] = webClientAssembly:GetType("ILLiad_WebClient_Override.ILLiadWebClientOverride");

local BASE_SOAP_XML = "<?xml version=\"1.0\" encoding=\"utf-8\"?><soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\"><soap:Header><UserCredentials xmlns=\"http://reprintsdesk.com/webservices/\"><UserName></UserName><Password></Password></UserCredentials></soap:Header><soap:Body></soap:Body></soap:Envelope>";
local REPRINTS_DESK_URL_STAGING = "https://www.stg.reprintsdesk.com/webservice/main.asmx";
local REPRINTS_DESK_URL_PRODUCTION = "https://www.reprintsdesk.com/webservice/main.asmx";
local REPRINTS_DESK_NAMESPACE = "http://reprintsdesk.com/webservices/";

local REPRINTS_DESK_SOAP_ACTION_ORDER_GET_PRICE_ESTIMATE_2 = "http://reprintsdesk.com/webservices/Order_GetPriceEstimate2";
local REPRINTS_DESK_SOAP_ACTION_ORDER_PLACE_ORDER_2 = "http://reprintsdesk.com/webservices/Order_PlaceOrder2";

local function CreateDefaultNamespaceManager(document)
	log:Debug("[ReprintsDeskService.CreateDefaultNamespaceManager]");

	local namespaceManager = types["System.Xml.XmlNamespaceManager"](document.NameTable);
	namespaceManager:AddNamespace("soap", "http://schemas.xmlsoap.org/soap/envelope/");
	namespaceManager:AddNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance");
	namespaceManager:AddNamespace("xsd", "http://www.w3.org/2001/XMLSchema");
	namespaceManager:AddNamespace("rd", REPRINTS_DESK_NAMESPACE);

	return namespaceManager;
end

local function CreateBaseDocument()
	local document = types["System.Xml.XmlDocument"]();
	document:LoadXml(BASE_SOAP_XML);

	local namespaceManager = CreateDefaultNamespaceManager(document);

	-- Initialize user credentials
	local userCredentialsNode = document:SelectSingleNode("/soap:Envelope/soap:Header/rd:UserCredentials", namespaceManager);

	local usernameNode = userCredentialsNode:SelectSingleNode("rd:UserName", namespaceManager);
	usernameNode.InnerText = ReprintsDeskService.Username;

	local passwordNode = userCredentialsNode:SelectSingleNode("rd:Password", namespaceManager);
	passwordNode.InnerText = ReprintsDeskService.Password;

	return document, namespaceManager;
end

local function CreateElement(document, name, namespace, parentNode)
	local node;

	if (namespace) then
		node = document:CreateElement(name, namespace);
	else
		node = document:CreateElement(name);
	end

	if (parentNode) then
		parentNode:AppendChild(node);
	end

	return node;
end

local function CreateAttribute(document, name, value, parentNode)
	local attribute = document:CreateAttribute(name);
	attribute.Value = (value or "");
	parentNode:SetAttributeNode(attribute);

	return attribute;
end

local function CreateTextNode(document, name, namespace, value, parentNode)
	local node = CreateElement(document, name, namespace, parentNode);
	node:AppendChild(document:CreateTextNode(value or ""));

	return node;
end

local function CreateCDataNode(document, name, namespace, value, parentNode)
	local node = CreateElement(document, name, namespace, parentNode);
	node:AppendChild(document:CreateCDataSection(value or ""));

	return node;
end

local function ParseSoapError(document, namespaceManager)
	log:Debug("[ReprintsDeskService.ReadSoapError]");

	local soapError = {};
	soapError.Code = "";
	soapError.Message = "";

	local faultNode = document:SelectSingleNode("/soap:Envelope/soap:Body/soap:Fault", namespaceManager);

	if (faultNode) then
		local faultCodeNode = faultNode:get_Item("faultcode");
		local faultStringNode = faultNode:get_Item("faultstring");

		if (faultCodeNode) then
			soapError.Code = faultCodeNode.InnerText;
		end

		if (faultStringNode) then
			soapError.Message = faultStringNode.InnerText;
		end
	else
		log:Debug("Could not locate fault node");
	end

	return soapError;
end

local function ParseOrderEstimate(document)
	log:Debug("[ReprintsDeskService.ParseOrderEstimate]");

	if (document == nil) then
		return nil;
	end

	local result = {};
	result.Error = nil;

	local namespaceManager = CreateDefaultNamespaceManager(document);

	local outputNode = document:SelectSingleNode("/soap:Envelope/soap:Body/rd:Order_GetPriceEstimate2Response/rd:xmlOutput/output", namespaceManager);

	if (outputNode) then
		log:Debug("Parsing output node");
		local standardNumberNode = outputNode:get_Item("standardNumber");
		local yearNode = outputNode:get_Item("year");
		local totalPagesNode = outputNode:get_Item("totalpages");
		local priceTypeNode = outputNode:get_Item("pricetypeid");
		local serviceChargeNode = outputNode:get_Item("servicecharge");
		local copyrightChargeNode = outputNode:get_Item("copyrightcharge");
		local disclaimerNode = outputNode:get_Item("disclaimer");

		-- Construct result
		if (standardNumberNode) then
			result.StandardNumber = standardNumberNode.InnerText;
		end

		if (yearNode) then
			result.Year = tonumber(yearNode.InnerText);
		end

		if (totalPagesNode) then
			result.TotalPages = tonumber(totalPagesNode.InnerText);
		end

		if (priceTypeNode) then
			result.PriceType = priceTypeNode.InnerText;
		end

		if (serviceChargeNode) then
			result.ServiceCharge = tonumber(serviceChargeNode.InnerText);
		end

		if (copyrightChargeNode) then
			result.CopyrightCharge = tonumber(copyrightChargeNode.InnerText);
		end

		if (disclaimerNode) then
			result.Disclaimer = disclaimerNode.InnerText;
		end
	else
		log:Debug("Parsing soap error");
		result.Error = ParseSoapError(document, namespaceManager);
	end

	return result;
end

local function ParseOrder(document)
	log:Debug("[ReprintsDeskService.ParseOrder]");

	if (document == nil) then
		return nil;
	end

	local result = {};
	result.Error = nil;

	local namespaceManager = CreateDefaultNamespaceManager(document);

	local responseNode = document:SelectSingleNode("/soap:Envelope/soap:Body/rd:Order_PlaceOrder2Response", namespaceManager);

	if (responseNode) then
		local resultNode = responseNode:get_Item("Order_PlaceOrder2Result");
		local orderIDNode = responseNode:get_Item("orderID");
		local rndIDNode = responseNode:get_Item("rndID");

		result.OrderResult = tonumber(resultNode.InnerText);
		result.OrderID = tonumber(orderIDNode.InnerText);
		result.RndID = tonumber(rndIDNode.InnerText);
	else
		result.Error = ParseSoapError(document, namespaceManager);
	end

	return result;
end

local function ReadResponse(responseString)
	log:Debug("[ReprintsDeskService.ReadResponse]");

	if (responseString and #responseString > 0) then
		Utility.Log(responseString, false);

		local responseDocument = types["System.Xml.XmlDocument"]();

		local documentLoaded = pcall(function ()
			responseDocument:LoadXml(responseString);
		end);

		if (documentLoaded) then
			return responseDocument;
		else
			log:Debug("Unable to load response content as XML");
			return nil;
		end
	else
		log:Debug("Unable to read response content");
	end

	return nil;
end

local function UploadStringCall(client, uri, doc)
	return client:UploadStringOverride(uri, doc);
end

function HandleUploadStringFailure(errorname)
	Utility.Log("[HandleUploadStringFailure]", false);

	local webExceptionResponse = errorname.InnerException.Response;
	if webExceptionResponse ~= nil then
		local responseStream = webExceptionResponse:GetResponseStream();
		if responseStream ~= nil then
			local responseStreamReader = types["StreamReader"](webExceptionResponse:GetResponseStream());
			if responseStreamReader ~= nil then
				responseString = responseStreamReader:ReadToEnd();
				if responseString ~= nil then
					--Convert the response into an XML document and parse the Error
					local responseDoc = ReadResponse(responseString);
					local namespaceManager = CreateDefaultNamespaceManager(responseDoc);
					local returnError = ParseSoapError(responseDoc, namespaceManager);

					return '['..returnError.Message..']';
				end
			end
		end
	end
end

local function SendRequest(request)
	log:Debug("[ReprintsDeskService.SendRequest]");
	Utility.Log(request, true);

	if (client == nil) then
		log:Debug("Creating web client");

		client = types["System.Activator"].CreateInstance(types["ILLiad-WebClient-Override.ILLiadWebClientOverride"]);
		client.Encoding = types["System.Text.Encoding"].UTF8;
	end

	client.Headers:Clear();
	client.Headers:Add("Accept", "text/xml");
	client.Headers:Add("Content-Type", "text/xml;charset=\"utf-8\"");
	client.Headers:Add("SOAPAction", request.Action);
	client.Headers:Add("User-Agent", (ReprintsDeskServiceInternal.UserAgent or "Reprints-Desk-Addon"));
	client.Headers:Add("Cache-Control", "no-cache");

	local uri;

	if (ReprintsDeskServiceInternal.IsProduction) then
		log:Debug("Using production URL");
		uri = types["System.Uri"](REPRINTS_DESK_URL_PRODUCTION);
	else
		log:Debug("Using staging URL");
		uri = types["System.Uri"](REPRINTS_DESK_URL_STAGING);
	end

	local doc = request.RequestDocument.OuterXml;

	log:Debug("ReprintsDeskService: Sending request");
	Utility.Log(doc, false);

	local success, serverResult = pcall(UploadStringCall, client, uri, doc);

	client:Dispose();

	log:Debug("ReprintsDeskService: Response Received");
	local result = {};
	Utility.Log(serverResult, false);

	if not success then
		log:Error("ReprintsDeskService: Error sending request");
		result.Error =  HandleUploadStringFailure(serverResult);
		result.Input = request.Input;
		Utility.Log(result.Error, false);
	else
		log:Debug("ReprintsDeskService: Request sent successful");

		result.ResponseDocument = ReadResponse(serverResult);
		result.Input = request.Input;
		result.Action = request.Action;
		result.RequestDocument = request.RequestDocument;

		Utility.Log(result.ResponseDocument, false);
	
		if (request.Action == REPRINTS_DESK_SOAP_ACTION_ORDER_GET_PRICE_ESTIMATE_2) then
			log:Debug("Order_GetPriceEstimate2 response");	
			result.Estimate = ParseOrderEstimate(result.ResponseDocument);
		elseif (request.Action == REPRINTS_DESK_SOAP_ACTION_ORDER_PLACE_ORDER_2) then
			log:Debug("Order_Place_Order2 response");
			result.Order = ParseOrder(result.ResponseDocument);	
		end
	end

	return result;
end

local function BuildOrderEstimate(input)
	log:Debug("[ReprintsDeskService.BuildOrderEstimate]");

	local document, namespaceManager = CreateBaseDocument();
	local bodyNode = document:SelectSingleNode("/soap:Envelope/soap:Body", namespaceManager);

	-- Order_GetPriceEstimate2
	local orderNode = CreateElement(document, "Order_GetPriceEstimate2", REPRINTS_DESK_NAMESPACE, bodyNode);
	local xmlInputNode = CreateElement(document, "xmlInput", REPRINTS_DESK_NAMESPACE, orderNode);
	local inputNode = CreateElement(document, "input", nil, xmlInputNode);

	CreateAttribute(document, "schemaversionid", "1", inputNode);
	CreateTextNode(document, "standardnumber", nil, input.StandardNumber, inputNode);
	CreateTextNode(document, "year", nil, input.Year, inputNode);
	CreateTextNode(document, "totalpages", nil, input.TotalPages, inputNode);
	CreateTextNode(document, "pricetypeid", nil, input.PriceType, inputNode);

	return document;
end

local function GetOrderEstimate(input)
	log:Debug("[ReprintsDeskService.GetOrderEstimate]");
	Utility.Log(input, true);

	local request = {};
	request.Action = REPRINTS_DESK_SOAP_ACTION_ORDER_GET_PRICE_ESTIMATE_2;
	request.Input = input;
	request.RequestDocument = BuildOrderEstimate(input);

	return SendRequest(request);
end

local function BuildOrder(input)
	log:Debug("[ReprintsDeskService.BuildOrder]");

	local document, namespaceManager = CreateBaseDocument();
	local bodyNode = document:SelectSingleNode("/soap:Envelope/soap:Body", namespaceManager);

	-- Order_PlaceOrder2
	local placeOrderNode = CreateElement(document, "Order_PlaceOrder2", REPRINTS_DESK_NAMESPACE, bodyNode);
	local xmlNode = CreateElement(document, "xmlNode", REPRINTS_DESK_NAMESPACE, placeOrderNode);
	local orderNode = CreateElement(document, "order", nil, xmlNode);

	-- Detail
	local orderDetailNode = CreateElement(document, "orderdetail", nil, orderNode);

	CreateTextNode(document, "ordertypeid", nil, input.Detail.OrderTypeID, orderDetailNode);
	CreateTextNode(document, "deliverymethodid", nil, input.Detail.DeliveryMethodID, orderDetailNode);
	CreateCDataNode(document, "comment", nil, input.Detail.Comment, orderDetailNode);
	CreateCDataNode(document, "aulast", nil, input.Detail.AuthorLast, orderDetailNode);
	CreateCDataNode(document, "aufirst", nil, input.Detail.AuthorFirst, orderDetailNode);
	CreateCDataNode(document, "issn", nil, input.Detail.ISSN, orderDetailNode);
	CreateCDataNode(document, "eissn", nil, input.Detail.EISSN, orderDetailNode);
	CreateCDataNode(document, "isbn", nil, input.Detail.ISBN, orderDetailNode);
	CreateCDataNode(document, "title", nil, input.Detail.Title, orderDetailNode);
	CreateCDataNode(document, "atitle", nil, input.Detail.ArticleTitle, orderDetailNode);
	CreateCDataNode(document, "volume", nil, input.Detail.Volume, orderDetailNode);
	CreateCDataNode(document, "issue", nil, input.Detail.Issue, orderDetailNode);
	CreateCDataNode(document, "spage", nil, input.Detail.StartPage, orderDetailNode);
	CreateCDataNode(document, "epage", nil, input.Detail.EndPage, orderDetailNode);
	CreateCDataNode(document, "pages", nil, input.Detail.Pages, orderDetailNode);
	CreateCDataNode(document, "date", nil, input.Detail.Date, orderDetailNode);
	if(input.Detail.DOI ~= nil and input.Detail.DOI ~= "") then
		CreateCDataNode(document, "doi", nil, input.Detail.DOI, orderDetailNode);
	end

	-- User
	local userNode = CreateElement(document, "user", nil, orderNode);

	CreateCDataNode(document, "username", nil, input.User.Username, userNode);
	CreateCDataNode(document, "email", nil, input.User.Email, userNode);
	CreateCDataNode(document, "billingreference", nil, input.User.BillingReference, userNode);
	CreateCDataNode(document, "firstname", nil, input.User.FirstName, userNode);
	CreateCDataNode(document, "lastname", nil, input.User.LastName, userNode);

	-- Delivery Profile
	local deliveryProfileNode = CreateElement(document, "deliveryprofile", nil, orderNode);

	CreateCDataNode(document, "firstname", nil, input.DeliveryProfile.FirstName, deliveryProfileNode);
	CreateCDataNode(document, "lastname", nil, input.DeliveryProfile.LastName, deliveryProfileNode);
	CreateCDataNode(document, "companyname", nil, input.DeliveryProfile.CompanyName, deliveryProfileNode);
	CreateCDataNode(document, "address1", nil, input.DeliveryProfile.Address1, deliveryProfileNode);
	CreateCDataNode(document, "address2", nil, input.DeliveryProfile.Address2, deliveryProfileNode);
	CreateCDataNode(document, "city", nil, input.DeliveryProfile.City, deliveryProfileNode);
	CreateCDataNode(document, "statecode", nil, input.DeliveryProfile.StateCode, deliveryProfileNode);
	CreateCDataNode(document, "statename", nil, input.DeliveryProfile.StateName, deliveryProfileNode);
	CreateCDataNode(document, "zip", nil, input.DeliveryProfile.Zip, deliveryProfileNode);
	CreateCDataNode(document, "countrycode", nil, input.DeliveryProfile.CountryCode, deliveryProfileNode);
	CreateCDataNode(document, "email", nil, input.DeliveryProfile.Email, deliveryProfileNode);
	CreateCDataNode(document, "phone", nil, input.DeliveryProfile.Phone, deliveryProfileNode);
	CreateCDataNode(document, "fax", nil, input.DeliveryProfile.Fax, deliveryProfileNode);

	-- Processing Instructions
	local processingInstructionsNode = CreateElement(document, "processinginstructions", nil, orderNode);

	for index, value in ipairs(input.ProcessingInstructions) do
		local node = CreateElement(document, "processinginstruction", nil, processingInstructionsNode);
		CreateAttribute(document, "id", index, node);
		CreateAttribute(document, "valueid", value, node);
	end

	-- Customer References
	local customerReferencesNode = CreateElement(document, "customerreferences", nil, orderNode);

	for index, value in ipairs(input.CustomerReferences) do
		local node = CreateCDataNode(document, "customerreference", nil, value, customerReferencesNode);
		CreateAttribute(document, "id", index, node);
	end

	return document;
end

local function PlaceOrder(input)
	log:Debug("[ReprintsDeskService.PlaceOrder]");
	Utility.Log(input, true);

	local request = {};
	request.Action = REPRINTS_DESK_SOAP_ACTION_ORDER_PLACE_ORDER_2;
	request.Input = input;
	request.RequestDocument = BuildOrder(input);

	return SendRequest(request);
end

-- Exports

ReprintsDeskService.GetOrderEstimate = GetOrderEstimate;
ReprintsDeskService.PlaceOrder = PlaceOrder;
